# -*- coding: utf-8 -*-
import pytest
from autoextract.request import (
    Request,
    ArticleRequest,
    ProductRequest,
    JobPostingRequest
)


def test_request():
    assert Request("http://example.com", pageType="foobar").as_dict() == {
        'url': 'http://example.com',
        'pageType': 'foobar',
        'articleBodyRaw': False,
    }


@pytest.mark.parametrize("cls,page_type", [
    [ArticleRequest, "article"],
    [ProductRequest, "product"],
    [JobPostingRequest, "jobPosting"]
])
def test_request_types(cls, page_type):
    req = cls("http://example.com")
    qdict = req.as_dict()
    assert qdict['pageType'] == page_type


def test_request_as_dict():
    request = Request("example.com", "myPageType")
    assert request.url == "example.com"
    assert request.pageType == "myPageType"
    assert request.fullHtml is None
    assert request.as_dict() == {
        'url': 'example.com',
        'pageType': 'myPageType',
        'articleBodyRaw': False,
    }

    request = Request("example.com", "myPageType", fullHtml=True)
    assert request.fullHtml is True
    assert request.as_dict() == {
        'url': 'example.com',
        'pageType': 'myPageType',
        'fullHtml': True,
        'articleBodyRaw': False,
    }

    request = Request("example.com", "myPageType", meta='foo bar', extra=dict(foo="bar"))
    assert request.url == "example.com"
    assert request.pageType == "myPageType"
    assert request.fullHtml is None
    assert request.as_dict() == {
        'url': 'example.com',
        'pageType': 'myPageType',
        'articleBodyRaw': False,
        'meta': 'foo bar',
        'foo': 'bar',
    }



# -*- coding: utf-8 -*-
from typing import List, Dict, Any, Union, Optional
import attr


QueryDict = Dict[str, Any]


@attr.s
class Request:
    """
    A single request data for AutoExtract.
    See https://doc.scrapinghub.com/autoextract.html#requests

    Note that `articleBodyRaw` is set to false by default here; API itself
    defaults to true. Set `articleBodyRaw=None` to remove `articleBodyRaw`
    parameter from the request and use server default.

    You can override API params passing a dictionary with extra data using the
    `extra` argument. Note that it will overwrite any previous configuration
    made using standard attributes like `articleBodyRaw` and `fullHtml`.
    """
    url = attr.ib()  # type: str
    pageType = attr.ib()  # type: str

    meta = attr.ib(default=None)  # type: Optional[str]
    articleBodyRaw = attr.ib(default=False)  # type: Optional[bool]
    fullHtml = attr.ib(default=None)  # type: Optional[bool]

    extra = attr.ib(default=None)  # type: Optional[dict]

    def as_dict(self) -> QueryDict:
        d = attr.asdict(self)
        d.update(**self.extra or {})
        del d['extra']
        return {key: value for key, value in d.items() if value is not None}


Query = Union[List[Request], List[QueryDict]]


@attr.s
class ArticleRequest(Request):
    pageType = attr.ib(default='article')


@attr.s
class ProductRequest(Request):
    pageType = attr.ib(default='product')


@attr.s
class JobPostingRequest(Request):
    pageType = attr.ib(default='jobPosting')


def query_as_dict_list(query: Query) -> List[QueryDict]:
    return [
        request.as_dict() if isinstance(request, Request) else request
        for request in query
    ]



import pytest

from autoextract.aio.client import RequestProcessor
from autoextract.aio.errors import QueryError


def test_request_processor_without_retries():
    # Given an initial query with two items
    initial_query = [
        {
            "url": "https://example.org/first",
            "pageType": "article",
        },
        {
            "url": "https://example.org/second",
            "pageType": "article",
        },
    ]

    # Initialize our request processor with this query
    request_processor = RequestProcessor(query=initial_query)

    # Since we haven't processed any response yet,
    # our request processor should have:
    #
    # Empty results
    assert request_processor.get_latest_results() == []
    # Pending queries list equal to our initial query
    assert request_processor.pending_queries == initial_query

    # Given our first response with a success and an error
    first_response = [
        {
            "query": {
                "id": "1",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/first",
                    "pageType": "article"
                }
            },
            "article": {
                "name": "My first article",
            },
        },
        {
            "query": {
                "id": "2",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/second",
                    "pageType": "article"
                }
            },
            "error": "Proxy error: internal_error",
        },
    ]

    # Process this response
    results = request_processor.process_results(first_response)

    # Results should be equal to our response
    assert results == first_response
    # Latest results should be equal to our response
    assert request_processor.get_latest_results() == results
    # Our pending queries list should be empty
    # because retry is disabled
    assert request_processor.pending_queries == []


def test_request_processor_with_not_retriable_error():
    # Given an initial query with two items
    initial_query = [
        {
            "url": "https://example.org/first",
            "pageType": "article",
        },
        {
            "url": "https://example.org/second",
            "pageType": "article",
        },
    ]

    # Initialize our request processor with this query
    request_processor = RequestProcessor(query=initial_query, max_retries=3)

    # Since we haven't processed any response yet,
    # our request processor should have:
    #
    # Empty results
    assert request_processor.get_latest_results() == []
    # Pending queries list equal to our initial query
    assert request_processor.pending_queries == initial_query

    # Given our first response with a success and an error
    first_response = [
        {
            "query": {
                "id": "1",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/first",
                    "pageType": "article"
                }
            },
            "article": {
                "name": "My first article",
            },
        },
        {
            "query": {
                "id": "2",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/second",
                    "pageType": "article"
                }
            },
            "error": "Downloader error: http404",
        },
    ]

    # When processing this response, no QueryError should be raised
    # because the 404 error is not retriable
    results = request_processor.process_results(first_response)

    # Results should be equal to our response
    assert results == first_response
    # Latest results should be equal to our response
    assert request_processor.get_latest_results() == results
    # Our pending queries list should be empty
    # because the errors found are not retriable
    assert request_processor.pending_queries == []


def test_request_processor_with_retries():
    # Given an initial query with two items
    initial_query = [
        {
            "url": "https://example.org/first",
            "pageType": "article",
        },
        {
            "url": "https://example.org/second",
            "pageType": "article",
        },
    ]

    # Initialize our request processor with this query
    request_processor = RequestProcessor(query=initial_query, max_retries=3)

    # Since we haven't processed any response yet,
    # our request processor should have:
    #
    # Empty results
    assert request_processor.get_latest_results() == []
    # Pending queries list equal to our initial query
    assert request_processor.pending_queries == initial_query

    # Given our first response with a success and an error
    first_response = [
        {
            "query": {
                "id": "1",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/first",
                    "pageType": "article"
                }
            },
            "article": {
                "name": "My first article",
            },
        },
        {
            "query": {
                "id": "2",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/second",
                    "pageType": "article"
                }
            },
            "error": "Proxy error: internal_error",
        },
    ]

    # An exception should be raised while processing this response
    with pytest.raises(QueryError):
        request_processor.process_results(first_response)

    # Latest results should be equal to our response
    assert request_processor.get_latest_results() == first_response
    # Our pending queries list should contain our query with error
    error_query = first_response[1]
    assert "error" in error_query
    user_query = error_query["query"]["userQuery"]
    assert request_processor.pending_queries == [user_query]

    # Given another error response for this missing query
    second_response = [
        {
            "query": {
                "id": "3",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/second",
                    "pageType": "article"
                }
            },
            "error": "Proxy error: internal_error",
        },
    ]

    # If we try to process this response,
    # it should raise an exception again
    with pytest.raises(QueryError):
        request_processor.process_results(second_response)

    # Latest results should be equal to a combination of our two responses:
    # successes from previous responses and most up to date errors.
    assert request_processor.get_latest_results() == [
        first_response[0],
        second_response[0],
    ]
    # Our pending queries list should contain our query with error
    error_query = second_response[0]
    assert "error" in error_query
    user_query = error_query["query"]["userQuery"]
    assert request_processor.pending_queries == [user_query]

    # Given another response, this time with a success
    third_response = [
        {
            "query": {
                "id": "4",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/second",
                    "pageType": "article"
                }
            },
            "article": {
                "name": "My second article",
            },
        },
    ]

    # If we try to process this response
    results = request_processor.process_results(third_response)
    # Results should be equal to a combination of success results
    # from previous and current requests
    assert results == [
        first_response[0],
        third_response[0],
    ]
    # Latest results should be equal to our response
    assert request_processor.get_latest_results() == results
    # Our pending queries list should be empty
    # because there's no additional query to be retried
    assert request_processor.pending_queries == []


@pytest.mark.parametrize("message, retry_seconds, domain_occupied", [
    ("Domain example.com is occupied, please retry in 23.5 seconds", 23.5, True,),
    ("Domain example.com is occupied, please retry in 14 seconds", 14.0, True,),
    ("Proxy error: timeout", 0.0, False,),
])
def test_request_processor_exception_priority(
        message, retry_seconds, domain_occupied):
    # Given an initial query with two items
    initial_query = [
        {
            "url": "https://example.org/first",
            "pageType": "article",
        },
        {
            "url": "https://example.org/second",
            "pageType": "article",
        },
    ]

    # Initialize our request processor with this query
    request_processor = RequestProcessor(query=initial_query, max_retries=3)

    # Since we haven't processed any response yet,
    # our request processor should have:
    #
    # Empty results
    assert request_processor.get_latest_results() == []
    # Pending queries list equal to our initial query
    assert request_processor.pending_queries == initial_query

    # Given our first response with two errors
    first_response = [
        {
            "query": {
                "id": "1",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/first",
                    "pageType": "article"
                }
            },
            "error": message,
        },
        {
            "query": {
                "id": "2",
                "domain": "example.org",
                "userQuery": {
                    "url": "https://example.org/second",
                    "pageType": "article"
                }
            },
            "error": "Proxy error: internal_error",
        },
    ]

    # If we try to process our response, a QueryError should be raised
    with pytest.raises(QueryError) as exc_info:
        request_processor.process_results(first_response)

    assert bool(exc_info.value.domain_occupied) is domain_occupied
    assert exc_info.value.retry_seconds == retry_seconds

    # The same thing should happen if the order of the queries is inverted
    first_response.reverse()
    with pytest.raises(QueryError) as exc_info:
        request_processor.process_results(first_response)

    assert bool(exc_info.value.domain_occupied) is domain_occupied
    assert exc_info.value.retry_seconds == retry_seconds



# -*- coding: utf-8 -*-
""" Basic command-line interface for AutoExtract. """
import argparse
import json
import sys
import asyncio
import logging
import random

import tqdm

from autoextract import Request
from autoextract.aio import (
    request_parallel_as_completed,
    create_session
)
from autoextract.stats import AggStats
from autoextract.aio.client import Result
from autoextract.constants import ENV_VARIABLE
from autoextract.request import Query


logger = logging.getLogger('autoextract')


async def run(query: Query, out, n_conn, batch_size, stop_on_errors=False,
              api_key=None, api_endpoint=None, max_query_error_retries=0):
    agg_stats = AggStats()
    async with create_session() as session:
        result_iter = request_parallel_as_completed(
            query=query,
            n_conn=n_conn,
            batch_size=batch_size,
            session=session,
            api_key=api_key,
            endpoint=api_endpoint,
            agg_stats=agg_stats,
            max_query_error_retries=max_query_error_retries
        )
        pbar = tqdm.tqdm(smoothing=0, leave=True, total=len(query), miniters=1,
                         unit="url")
        pbar.set_postfix_str(str(agg_stats))
        try:
            for fut in result_iter:
                try:
                    batch_result: Result = await fut
                    for res in batch_result:
                        json.dump(res, out, ensure_ascii=False)
                        out.write("\n")
                        out.flush()
                        pbar.update()
                except Exception as e:
                    if stop_on_errors:
                        raise
                    logger.error(str(e))
                finally:
                    pbar.set_postfix_str(str(agg_stats))
        finally:
            pbar.close()


def read_input(input_fp, intype, page_type):
    assert intype in {"txt", "jl", ""}
    if intype == "txt":
        urls = [u.strip() for u in input_fp.readlines() if u.strip()]
        query = [Request(url, pageType=page_type) for url in urls]
        return query
    elif intype == "jl":
        records = [
            json.loads(line.strip())
            for line in input_fp.readlines() if line.strip()
        ]
        for rec in records:
            rec.setdefault("pageType", page_type)
            if not isinstance(rec.get("meta", ""), (str, type(None))):
                raise TypeError("meta must be str or null, got {!r}".format(rec['meta']))
        return records


if __name__ == '__main__':
    """ Process urls from input file through AutoExtract """
    p = argparse.ArgumentParser(
        prog='python -m autoextract',
        description="""
        Process input URLs from a file using AutoExtract.
        """,
    )
    p.add_argument("input",
                   type=argparse.FileType("r", encoding='utf8'),
                   help="Input file with urls, url per line by default. The "
                        "Format can be changed using `--intype` argument.")
    p.add_argument("--intype", default="txt", choices=["txt", "jl"], 
                   help='Type of the input file (default: %(default)s). '
                        'Allowed values are "txt": input should be one '
                        'URL per line, and "jl": input should be a jsonlines '
                        'file, with {"url": "...", "meta": ...,} dicts;'
                        'see https://doc.scrapinghub.com/autoextract.html#requests'
                        'for the data format description.')
    p.add_argument("--output", "-o",
                   default=sys.stdout,
                   type=argparse.FileType("w", encoding='utf8'),
                   help=".jsonlines file to store extracted data. "
                        "By default, results are printed to stdout.")
    p.add_argument("--n-conn", type=int, default=20,
                   help="number of connections to the API server "
                        "(default: %(default)s)")
    p.add_argument("--batch-size", type=int, default=2,
                   help="batch size (default: %(default)s)")
    p.add_argument("--page-type", "-t", default="article",
                   help="type of the pages in the input file, "
                        "e.g. article, product, jobPosting "
                        "(default: %(default)s)")
    p.add_argument("--api-key",
                   help="Scrapinghub AutoExtract API key. "
                        "You can also set %s environment variable instead "
                        "of using this option." % ENV_VARIABLE)
    p.add_argument("--api-endpoint",
                   help="Scrapinghub AutoExtract API endpoint.")
    p.add_argument("--loglevel", "-L", default="INFO",
                   choices=["DEBUG", "INFO", "WARNING", "ERROR"],
                   help="log level")
    p.add_argument("--shuffle", help="Shuffle input URLs", action="store_true")
    p.add_argument("--max-query-error-retries", type=int, default=0,
                   help="Max number of Query-level error retries. "
                        "Enable Query-level error retries to increase the "
                        "success rate at the cost of more requests being "
                        "performed. It is recommended if you are interested "
                        "in a higher success rate.")
    args = p.parse_args()
    logging.basicConfig(level=getattr(logging, args.loglevel))

    query = read_input(args.input, args.intype, args.page_type)
    if args.shuffle:
        random.shuffle(query)

    logger.info(f"Loaded {len(query)} urls from {args.input.name}; shuffled: {args.shuffle}")
    logger.info(f"Running AutoExtract (connections: {args.n_conn}, "
                f"batch size: {args.batch_size}, page type: {args.page_type})")

    loop = asyncio.get_event_loop()
    coro = run(query,
               out=args.output,
               n_conn=args.n_conn,
               batch_size=args.batch_size,
               stop_on_errors=False,
               api_key=args.api_key,
               api_endpoint=args.api_endpoint,
               max_query_error_retries=args.max_query_error_retries)
    loop.run_until_complete(coro)
    loop.close()



# -*- coding: utf-8 -*-
"""
aiohttp Scrapinghub AutoExtract API client.
"""
import asyncio
import time
import warnings
from typing import Optional, Dict, List, Iterator
from functools import partial

import aiohttp

from autoextract.constants import API_ENDPOINT, API_TIMEOUT
from autoextract.apikey import get_apikey
from autoextract.utils import chunks, user_agent
from autoextract.request import Query, query_as_dict_list
from autoextract.stats import ResponseStats, AggStats
from .retry import autoextract_retry, QueryRetryError
from .errors import RequestError, QueryError


AIO_API_TIMEOUT = aiohttp.ClientTimeout(total=API_TIMEOUT + 60,
                                        sock_read=API_TIMEOUT + 30,
                                        sock_connect=10)


def create_session(**kwargs) -> aiohttp.ClientSession:
    """ Create a session with parameters suited for AutoExtract """
    kwargs.setdefault('timeout', AIO_API_TIMEOUT)
    return aiohttp.ClientSession(**kwargs)


class Result(List[Dict]):
    retry_stats: Optional[Dict] = None
    response_stats: Optional[List[ResponseStats]] = None


class RequestProcessor:
    """Help keeping track of query results and errors between retries.

    After initializing your Request Processor,
    you may use it for just a single or for multiple requests.

    This class is especially useful because it stores
    successful queries to avoid repeating them when retrying requests.
    """

    def __init__(self, query: Query, max_retries: int = 0):
        """Reset temporary data structures and initialize them"""
        self._reset()
        self.pending_queries = query_as_dict_list(query)
        self._max_retries = max_retries
        self._complete_queries: List[Dict] = list()

    def _reset(self):
        """Clear temporary variables between retries"""
        self.pending_queries: List[Dict] = list()
        self._retriable_queries: List[Dict] = list()
        self._retriable_query_exceptions: List[Dict] = list()

    def _enqueue_error(self, query_result, query_exception):
        """Enqueue Query-level error.

        Enqueued errors could be:
            - used in combination with successes with `get_latest_results`
            - retried using `pending_requests`
        """
        self._retriable_queries.append(query_result)
        self._retriable_query_exceptions.append(query_exception)

        user_query = query_result["query"]["userQuery"]
        self.pending_queries.append(user_query)

    def get_latest_results(self):
        """Get latest results (errors + successes).

        This method could be used to retrieve results
        when an exception is raised while processing results.
        """
        return self._complete_queries + self._retriable_queries

    def process_results(self, query_results):
        """Process query results.

        Return successful queries and also failed ones.

        If `self._max_retries` is greater than 0,
        this method might raise a `QueryError` exception.

        If multiple `QueryError` exceptions are parsed,
        the one with the longest timeout is raised.

        Successful requests are saved in `self._complete_queries`
        among with errors that cannot be retried,
        and they are kept between executions
        while retriable failures are saved in `self._retriable_queries`.

        Queries saved in `self._retriable_queries` are moved to
        `self.pending_queries` between executions.
        You can use the first or the n-th result:

            - You can get all queries successfully responded in the first try.
            - You can get all queries successfully in the n-th try.
            - You may stop with a partial number of successful queries.
        """
        self._reset()
        for query_result in query_results:
            if self._max_retries and "error" in query_result:
                query_exception = QueryError.from_query_result(
                    query_result, self._max_retries)
                if query_exception.retriable:
                    self._enqueue_error(query_result, query_exception)
                    continue

            self._complete_queries.append(query_result)

        if self._retriable_query_exceptions:
            # Prioritize exceptions that have retry seconds defined
            # and get the one with the longest timeout value
            exception_with_longest_timeout = max(
                self._retriable_query_exceptions,
                key=lambda exc: exc.retry_seconds
            )
            raise exception_with_longest_timeout

        return self.get_latest_results()


async def request_raw(query: Query,
                      api_key: Optional[str] = None,
                      endpoint: Optional[str] = None,
                      *,
                      handle_retries: bool = True,
                      max_query_error_retries: int = 0,
                      session: Optional[aiohttp.ClientSession] = None,
                      agg_stats: AggStats = None,
                      ) -> Result:
    """ Send a request to Scrapinghub AutoExtract API.

    ``query`` is a list of dicts or Request objects, as
    described in the API docs
    (see https://doc.scrapinghub.com/autoextract.html).

    ``api_key`` is your AutoExtract API key. If not set, it is
    taken from SCRAPINGHUB_AUTOEXTRACT_KEY environment variable.

    ``session`` is an optional aiohttp.ClientSession object;
    use it to enable HTTP Keep-Alive.

    This function retries http 429 errors and network errors by default;
    this allows to handle server-side throttling properly.
    Use ``handle_retries=False`` if you want to disable this behavior
    (e.g. to implement it yourself).

    When handle_retries is True, this function can raise

    1) autoextract.errors.RequestError,
       if there is a Request-level error returned by the API
       which is not a throttling response
       (e.g. it can be raised for incorrect request).
    2) tenacity.RetryError,
       if a network-related error persists for a long time,
       over the allowed time period.

    Throttling errors are retried indefinitely when handle_retries is True.

    When ``handle_retries=True``, we could also retry Query-level errors.
    Use ``max_query_error_retries > 0`` if you want to to enable this behavior.

    ``agg_stats`` argument allows to keep track of various stats; pass an
    ``AggStats`` instance, and it'll be updated.

    See :func:`request_parallel_as_completed` for a more high-level
    interface to send requests in parallel.
    """
    endpoint = API_ENDPOINT if endpoint is None else endpoint

    if agg_stats is None:
        agg_stats = AggStats()  # dummy stats, to simplify code

    if max_query_error_retries and not handle_retries:
        warnings.warn(
            "You've specified a max number of Query-level error retries, "
            "but retries are disabled. Consider passing the handle_retries "
            "argument as True.",
            stacklevel=2
        )

    # Keep state between executions/retries
    request_processor = RequestProcessor(
        query=query,
        max_retries=max_query_error_retries if handle_retries else 0,
    )

    post = _post_func(session)
    auth = aiohttp.BasicAuth(get_apikey(api_key))
    headers = {'User-Agent': user_agent(aiohttp)}

    response_stats = []
    start_global = time.perf_counter()

    async def request():
        stats = ResponseStats.create(start_global)
        agg_stats.n_attempts += 1

        post_kwargs = dict(
            url=endpoint,
            json=request_processor.pending_queries,
            auth=auth,
            headers=headers,
        )

        try:
            async with post(**post_kwargs) as resp:
                stats.status = resp.status
                stats.record_connected(agg_stats)
                if resp.status >= 400:
                    content = await resp.read()
                    resp.release()
                    stats.record_read()
                    stats.error = content
                    if resp.status == 429:
                        agg_stats.n_429 += 1
                    else:
                        agg_stats.n_errors += 1
                    raise RequestError(
                        request_info=resp.request_info,
                        history=resp.history,
                        status=resp.status,
                        message=resp.reason,
                        headers=resp.headers,
                        response_content=content
                    )

                response = await resp.json()
                stats.record_read(agg_stats)
                return request_processor.process_results(response)
        except Exception as e:
            if not isinstance(e, RequestError):
                agg_stats.n_errors += 1
            raise
        finally:
            response_stats.append(stats)

    if handle_retries:
        # If handle_retries=True, the request method could raise
        # RetryError and QueryRetryError exceptions.
        #
        # These exceptions are raised when Tenacity is not able to
        # successfully retry failing requests.
        #
        # In addition to handle_retries=True, QueryRetryError also depends on
        # max_query_error_retries being greater than 0.
        request = autoextract_retry(request)

    try:
        # Try to make a batch request
        result = await request()
    except QueryRetryError:
        # If Tenacity fails to retry a QueryError because the max number of
        # retries or a timeout was reached, get latest results combining
        # error and successes and consider it as the final result.
        result = request_processor.get_latest_results()
    except Exception:
        agg_stats.n_fatal_errors += 1
        raise

    result = Result(result)
    result.response_stats = response_stats
    if handle_retries:
        result.retry_stats = request.retry.statistics  # type: ignore

    agg_stats.n_results += 1
    return result


def request_parallel_as_completed(query: Query,
                                  api_key: Optional[str] = None,
                                  *,
                                  endpoint: Optional[str] = None,
                                  session: Optional[aiohttp.ClientSession] = None,
                                  batch_size=1,
                                  n_conn=1,
                                  agg_stats: AggStats = None,
                                  max_query_error_retries=0,
                                  ) -> Iterator[asyncio.Future]:
    """ Send multiple requests to AutoExtract API in parallel.
    Return an `asyncio.as_completed` iterator.

    ``query`` is a list of requests to process (autoextract.Request
    instances or dicts).

    ``api_key`` is your AutoExtract API key. If not set, it is
    taken from SCRAPINGHUB_AUTOEXTRACT_KEY environment variable.

    ``n_conn`` is a number of parallel connections to a server.
    ``batch_size`` is an amount of queries sent in a batch in each connection.
    Higher batch_size increase response time, but allows to achieve the same
    throughput with less connections to server.

    For example, if your API key has a limit of 3RPS, and average response
    time you observe for your websites is 10s, then to get to these
    3RPS you may set e.g. batch_size=2, n_conn=15 - this would allow to
    process 30 requests in parallel.

    ``session`` is an optional aiohttp.ClientSession object;
    use it to enable HTTP Keep-Alive.

    ``agg_stats`` argument allows to keep track of various stats; pass an
    ``AggStats`` instance, and it'll be updated.

    Use ``max_query_error_retries > 0`` if you want Query-level
    errors to be retried.
    """
    sem = asyncio.Semaphore(n_conn)

    async def _request(batch_query):
        async with sem:
            return await request_raw(batch_query,
                                     api_key=api_key,
                                     endpoint=endpoint,
                                     session=session,
                                     agg_stats=agg_stats,
                                     max_query_error_retries=max_query_error_retries,
                                     )

    batches = chunks(query, batch_size)
    return asyncio.as_completed([_request(batch) for batch in batches])


def _post_func(session):
    """ Return a function to send a POST request """
    if session is None:
        return partial(aiohttp.request,
                       method='POST',
                       timeout=AIO_API_TIMEOUT)
    else:
        return session.post



# -*- coding: utf-8 -*-
import pytest

from autoextract.aio.errors import QueryError


def get_query_error(message):
    return {
        "query": {
            "id": "1587642195276-9386233af6ce1b9f",
            "domain": "example.com",
            "userQuery": {
                "url": "http://www.example.com/this-page-does-not-exist",
                "pageType": "article"
            },
        },
        "error": message,
    }


@pytest.mark.parametrize("message, retriable", (
        ("Downloader error: http404", False),
        ("query timed out", True),
        ("Downloader error: No response (network5)", True),
        ("Downloader error: http50", True),
        ("Downloader error: GlobalTimeoutError", True),
        ("Proxy error: banned", True),
        ("Proxy error: internal_error", True),
        ("Proxy error: nxdomain", True),
        ("Proxy error: timeout", True),
        ("Proxy error: ssl_tunnel_error", True),
        ("Proxy error: msgtimeout", True),
        ("Proxy error: econnrefused", True),
))
def test_query_error(message, retriable):
    query_error = get_query_error(message)

    exc = QueryError.from_query_result(query_error)
    assert exc.query == query_error["query"]
    assert exc.message == query_error["error"]
    assert exc.retriable is retriable
    assert exc.domain_occupied is None
    assert str(exc) == f"QueryError: query={exc.query}, message={exc.message}, max_retries={exc.max_retries}"

    exc = QueryError(query=query_error["query"], message=query_error["error"])
    assert exc.query == query_error["query"]
    assert exc.message == query_error["error"]
    assert exc.retriable is retriable
    assert exc.domain_occupied is None
    assert str(exc) == f"QueryError: query={exc.query}, message={exc.message}, max_retries={exc.max_retries}"


@pytest.mark.parametrize("message, domain, retry_seconds", (
    ("Domain example.com is occupied, please retry in 23.5 seconds", "example.com", 23.5),
    ("domain example.net is occupied, please retry in 23.5 seconds", "example.net", 23.5),
    ("Domain example.com is occupied, please retry in 23 seconds", "example.com", 23),
    ("domain example.net is occupied, please retry in 23 seconds", "example.net", 23),
    ("Domain example.com is occupied, please retry in asd seconds", "example.com", 300),  # 5 minutes default
    ("domain example.net is occupied, please retry in xyz seconds", "example.net", 300),  # 5 minutes default
))
def test_domain_occupied_query_error(message, domain, retry_seconds):
    query_error = get_query_error(message)
    exc = QueryError.from_query_result(query_error)
    assert exc.domain_occupied.domain == domain
    assert exc.domain_occupied.retry_seconds == retry_seconds
    assert exc.retry_seconds == retry_seconds
    assert exc.retriable is True


@pytest.mark.parametrize("message", (
    "domain example.net is occupied, please retry in  seconds",
    "domain example.net is occupied, please retry in seconds",
    "foo bar",
))
def test_invalid_domain_occupied_query_error(message):
    query_error = get_query_error(message)
    exc = QueryError.from_query_result(query_error)
    assert exc.domain_occupied is None
    assert exc.retry_seconds == 0.0



# -*- coding: utf-8 -*-
import logging
import re
from typing import Optional

from aiohttp import ClientResponseError
from tenacity import RetryError

logger = logging.getLogger(__name__)


class DomainOccupied:

    DOMAIN_OCCUPIED_REGEX = re.compile(
        r".*domain (.+) is occupied, please retry in (.+) seconds.*",
        re.IGNORECASE
    )

    DEFAULT_RETRY_SECONDS = 5 * 60  # 5 minutes

    def __init__(self, domain: str, retry_seconds: float):
        self.domain = domain
        self.retry_seconds = retry_seconds

    @classmethod
    def from_message(cls, message: str) -> Optional["DomainOccupied"]:
        match = cls.DOMAIN_OCCUPIED_REGEX.match(message)
        if not match:
            return None

        domain = match.group(1)

        try:
            retry_seconds = float(match.group(2))
        except ValueError:
            logger.warning(
                f"Could not extract retry seconds "
                f"from Domain Occupied error message: {message}"
            )
            retry_seconds = cls.DEFAULT_RETRY_SECONDS

        return cls(domain=domain, retry_seconds=retry_seconds)


class RequestError(ClientResponseError):
    """ Exception which is raised when Request-level error is returned.
    In contrast with ClientResponseError, it allows to inspect response
    content.
    https://doc.scrapinghub.com/autoextract.html#request-level
    """
    def __init__(self, *args, **kwargs):
        self.response_content = kwargs.pop("response_content")
        super().__init__(*args, **kwargs)

    def __str__(self):
        return f"RequestError: {self.status}, message={self.message}, " \
               f"headers={self.headers}, body={self.response_content}"


class QueryError(Exception):
    """ Exception which is raised when a Query-level error is returned.
    https://doc.scrapinghub.com/autoextract.html#query-level
    """

    RETRIABLE_MESSAGES = {
        message.lower().strip()
        for message in [
            "query timed out",
            "Downloader error: No response (network5)",
            "Downloader error: http50",
            "Downloader error: GlobalTimeoutError",
            "Proxy error: banned",
            "Proxy error: internal_error",
            "Proxy error: nxdomain",
            "Proxy error: timeout",
            "Proxy error: ssl_tunnel_error",
            "Proxy error: msgtimeout",
            "Proxy error: econnrefused",
        ]
    }

    def __init__(self, query: dict, message: str, max_retries: int = 0):
        self.query = query
        self.message = message
        self.max_retries = max_retries
        self.domain_occupied = DomainOccupied.from_message(message)

    def __str__(self):
        return f"QueryError: query={self.query}, message={self.message}, " \
               f"max_retries={self.max_retries}"

    @classmethod
    def from_query_result(cls, query_result: dict, max_retries: int = 0):
        return cls(query=query_result["query"], message=query_result["error"],
                   max_retries=max_retries)

    @property
    def retriable(self) -> bool:
        if self.domain_occupied:
            return True

        return self.message.lower().strip() in self.RETRIABLE_MESSAGES

    @property
    def retry_seconds(self) -> float:
        if self.domain_occupied:
            return self.domain_occupied.retry_seconds

        return 0.0


class QueryRetryError(RetryError):
    """This exception is raised when Tenacity reaches the max retry count or
    timeouts when retrying Query-level errors (see :class:``.QueryError``).
    """
    pass



# -*- coding: utf-8 -*-
from typing import Optional
import functools
import time

import attr
from runstats import Statistics


def zero_on_division_error(meth):
    @functools.wraps(meth)
    def wrapper(*args, **kwargs):
        try:
            return meth(*args, **kwargs)
        except ZeroDivisionError:
            return 0
    return wrapper


class AggStats:
    def __init__(self):
        self.time_connect_stats = Statistics()
        self.time_total_stats = Statistics()
        self.n_results = 0
        self.n_fatal_errors = 0

        self.n_attempts = 0
        self.n_429 = 0
        self.n_errors = 0

    def __str__(self):
        return "conn:{:0.2f}s, resp:{:0.2f}s, throttle:{:.1%}, err:{}+{}({:.1%}) | {:.1%} success".format(
            self.time_connect_stats.mean(),
            self.time_total_stats.mean(),
            self.throttle_ratio(),
            self.n_errors - self.n_fatal_errors,
            self.n_fatal_errors,
            self.error_ratio(),
            self.success_ratio(),
        )

    @zero_on_division_error
    def throttle_ratio(self):
        return self.n_429 / self.n_attempts

    @zero_on_division_error
    def error_ratio(self):
        return self.n_errors / self.n_attempts

    @zero_on_division_error
    def success_ratio(self):
        return self.n_results / (self.n_results + self.n_fatal_errors)


@attr.s
class ResponseStats:
    _start = attr.ib(repr=False)  # type: float

    # Wait time, before this request is sent. Can be large in case of retries.
    time_delayed = attr.ib(default=None)  # type: Optional[float]

    # Time between sending a request and having a connection established
    time_connect = attr.ib(default=None)  # type: Optional[float]

    # Time to read & decode the response
    time_read = attr.ib(default=None)  # type: Optional[float]

    # Total time to process the response, excluding the wait time caused
    # by retries.
    time_total = attr.ib(default=None)  # type: Optional[float]

    # HTTP status code
    status = attr.ib(default=None)  # type: Optional[int]

    # response content, in case of error response
    error = attr.ib(default=None)  # type: Optional[bytes]

    @classmethod
    def create(cls, start_global):
        start = time.perf_counter()
        return cls(
            start=start,
            time_delayed=start - start_global,
        )

    def record_connected(self, agg_stats: AggStats):
        self.time_connect = time.perf_counter() - self._start
        agg_stats.time_connect_stats.push(self.time_connect)

    def record_read(self, agg_stats: Optional[AggStats]=None):
        now = time.perf_counter()
        self.time_total = now - self._start
        self.time_read = self.time_total - (self.time_connect or 0)
        if agg_stats:
            agg_stats.time_total_stats.push(self.time_total)
